package ports

import (
	"context"
	"fmt"

	"code.fbi.h-da.de/danet/gosdn/controller/customerrs"
	"code.fbi.h-da.de/danet/gosdn/controller/interfaces/networkelement"
	query "code.fbi.h-da.de/danet/gosdn/controller/store"

	"github.com/google/uuid"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
)

const storeName = "port-store.json"

// Store defines a PortStore interface.
type Store interface {
	Add(context.Context, Port) error
	Update(context.Context, Port) error
	Delete(context.Context, Port) error
	Get(context.Context, query.Query) (Port, error)
	GetAll(context.Context) ([]Port, error)
}

// DatabasePortStore is a database store for ports.
type DatabasePortStore struct {
	collection *mongo.Collection
}

// NewDatabasePortStore returns a PortStore.
func NewDatabasePortStore(db *mongo.Database) Store {
	collection := db.Collection(storeName)

	return &DatabasePortStore{
		collection: collection,
	}
}

// Get takes a Ports's UUID or name and returns the port.
func (s *DatabasePortStore) Get(ctx context.Context, query query.Query) (Port, error) {
	var loadedPort Port

	if query.ID.String() != "" {
		loadedPort, err := s.getByID(ctx, query.ID)
		if err != nil {
			return loadedPort, customerrs.CouldNotFindError{ID: query.ID, Name: query.Name}
		}

		return loadedPort, nil
	}

	loadedPort, err := s.getByName(ctx, query.Name)
	if err != nil {
		return loadedPort, customerrs.CouldNotFindError{ID: query.ID, Name: query.Name}
	}

	return loadedPort, nil
}

func (s *DatabasePortStore) getByID(ctx context.Context, idOfPort uuid.UUID) (loadedPort Port, err error) {
	idAsByteArray, _ := idOfPort.MarshalBinary()

	result := s.collection.FindOne(ctx, bson.D{primitive.E{Key: "_id", Value: idAsByteArray}})
	if result == nil {
		return loadedPort, customerrs.CouldNotFindError{ID: idOfPort}
	}

	err = result.Decode(&loadedPort)
	if err != nil {
		return loadedPort, customerrs.CouldNotMarshallError{Identifier: idOfPort, Type: loadedPort, Err: err}
	}

	return loadedPort, nil
}

func (s *DatabasePortStore) getByName(ctx context.Context, nameOfPort string) (loadedPort Port, err error) {
	result := s.collection.FindOne(ctx, bson.D{primitive.E{Key: "name", Value: nameOfPort}})
	if result == nil {
		return loadedPort, customerrs.CouldNotFindError{Name: nameOfPort}
	}

	err = result.Decode(&loadedPort)
	if err != nil {
		return loadedPort, customerrs.CouldNotMarshallError{Identifier: nameOfPort, Type: loadedPort, Err: err}
	}

	return loadedPort, nil
}

// GetAll returns all stored ports.
func (s *DatabasePortStore) GetAll(ctx context.Context) (loadedPorts []Port, err error) {
	cursor, err := s.collection.Find(ctx, bson.D{})
	if err != nil {
		return []Port{}, err
	}
	defer func() {
		if ferr := cursor.Close(ctx); ferr != nil {
			fErrString := ferr.Error()
			err = fmt.Errorf("InternalError=%w DeferError=%+s", err, fErrString)
		}
	}()

	err = cursor.All(ctx, &loadedPorts)
	if err != nil {
		return loadedPorts, customerrs.CouldNotMarshallError{Type: loadedPorts, Err: err}
	}

	return loadedPorts, nil
}

// Add adds a port to the port store.
func (s *DatabasePortStore) Add(ctx context.Context, port Port) (err error) {
	_, err = s.collection.
		InsertOne(ctx, port)
	if err != nil {
		return customerrs.CouldNotCreateError{Identifier: port.ID, Type: port, Err: err}
	}

	return nil
}

// Update updates a existing port.
func (s *DatabasePortStore) Update(ctx context.Context, port Port) (err error) {
	var updatedLoadedNetworkElement networkelement.LoadedNetworkElement

	update := bson.D{primitive.E{Key: "$set", Value: port}}

	upsert := false
	after := options.After
	opt := options.FindOneAndUpdateOptions{
		Upsert:         &upsert,
		ReturnDocument: &after,
	}

	err = s.collection.
		FindOneAndUpdate(
			ctx, bson.M{"_id": port.ID.String()}, update, &opt).
		Decode(&updatedLoadedNetworkElement)
	if err != nil {
		return customerrs.CouldNotUpdateError{Identifier: port.ID, Type: port, Err: err}
	}

	return nil
}

// Delete deletes a port from the port store.
func (s *DatabasePortStore) Delete(ctx context.Context, port Port) (err error) {
	_, err = s.collection.DeleteOne(ctx, bson.D{primitive.E{Key: port.ID.String()}})
	if err != nil {
		return err
	}

	return nil
}
